import pathlib

import click
import toml as toml_

from ._site import Site

from netlink.logging import logger
logger.set_level(logger.WARNING)


@click.command()
@click.option('-u', '--url', help='Sharepoint URL')
@click.option('-i', '--id', help='Client ID')
@click.option('-s', '--secret', help='Client Secret')
@click.option('-t', '--toml', type=click.Path(exists=True, dir_okay=False, path_type=pathlib.Path),  help="TOML file with 'url', 'id', and 'secret'")
@click.option('-f', '--fields', is_flag=True, help='include fields')
@click.option('--hidden', is_flag=True, help='include hidden lists')
@click.argument('name', nargs=-1)
def print_list_info(name, url, id, secret, toml, fields, hidden):
    """Print information about SharePoint List(s)

    If NAME is not provided, all lists are returned.

    """
    if toml is not None:
        with toml.open('r', encoding='utf-8-sig') as f:
            d = toml_.load(f)
        url, id, secret = d['url'], d['id'], d['secret']
    if url is None or id is None or secret is None:
        raise click.UsageError("Essential options missing. Either provide 'url', 'id', and 'secret', or a "
                               "respective 'toml'-file.")
    site = Site(url=url, client_id=id, client_secret=secret)
    for i in site.get_lists(hidden=hidden):
        if name:
            if i.title not in name:
                continue
        print(i.title)
        if fields:
            for j in site.get_list_columns(name=i.title, hidden=hidden):
                print(
                    f"    {j.title:25}  {j.type_as_string:10}  {j.internal_name}{'Id' if j.type_as_string in ('User', 'Lookup') else ''}")
            print()



import collections.abc

from office365.sharepoint.listitems.listitem import ListItem
from netlink.logging import logger


def item_factory(sharepoint_list):
    _len = len(sharepoint_list.sharepoint_map) + 1
    _iter = [i for i in sharepoint_list.python_map]
    _iter.insert(0, 'id')
    _iter = tuple(_iter)

    class Item(collections.abc.MutableMapping):
        _sharepoint_list = sharepoint_list

        def __init__(self, *args, **kwargs):
            if isinstance(args[0], ListItem):
                self._list_item: ListItem = args[0]
                self._data = dict(id=self._list_item.id)
                for sharepoint_column in self._sharepoint_list.sharepoint_map:
                    v = self._list_item.get_property(sharepoint_column)
                    if v:
                        python_column = self._sharepoint_list.sharepoint_map[sharepoint_column]
                        logger.trace(f'{sharepoint_column}: {v} -> {python_column}')
                        self._data[python_column] = v
            else:
                raise NotImplementedError
            self._dirty = set()

        def __len__(self) -> int:
            return _len

        def __iter__(self):
            return iter(_iter)

        def __getitem__(self, item):
            if item in _iter:
                return self._data.get(item, None)
            raise KeyError(item)

        def __setitem__(self, key, value):
            # id can only be changed when committed
            if key in self._sharepoint_list.python_map:
                if self._data.get(key, None) != value:
                    self._data[key] = value
                    self._dirty.add(key)
            else:
                raise KeyError(key)

        def __delitem__(self, key):
            raise NotImplementedError

        def __getattr__(self, item):
            try:
                return self[item]
            except KeyError:
                raise AttributeError(item)

        def commit(self, lazy: bool = False):
            if self._dirty:
                for i in self._dirty:
                    self._list_item.set_property(self._sharepoint_list.python_map[i], self._data[i]).update()
                self._dirty.clear()
                if not lazy:
                    self._list_item.execute_query()

        def __repr__(self):
            return f'{self._sharepoint_list.title}: {self._data}'

    return Item



from copy import deepcopy
from office365.sharepoint.client_context import ClientContext
from office365.runtime.auth.client_credential import ClientCredential

from netlink.logging import logger


class Site:
    _url = ''
    _id = ''
    _secret = ''

    def __init__(self, url: str = None, client_id: str = None, client_secret: str = None):
        self._context = ClientContext(url or self._url).with_credentials(
            ClientCredential(client_id=client_id or self._id, client_secret=client_secret or self._secret))
        logger.debug(f'Initialized connection Sharepoint as {self.url}')
        self._users = None
        self._lists = {}

    def get_list(self, name: str):
        if name not in self._lists:
            self._lists[name] = self._context.web.lists.get_by_title(name)
        return self._lists[name]

    @property
    def url(self):
        return self._context.base_url

    def get_lists(self, hidden: bool = False):
        if hidden:
            selector = lambda x: True
        else:
            selector = lambda x: x.properties['BaseTemplate'] == 100 and x.title not in ('TaxonomyHiddenList',)
        return [i for i in self._context.lists.get().execute_query() if selector(i)]

    def get_list_items(self, name):
        return self.get_list(name).items.get().execute_query()

    def commit(self):
        self._context.execute_batch()

    def get_list_columns(self, name, hidden=False):
        if not isinstance(name, str):
            name = name.title
        if hidden:
            selector = lambda x: True
        else:
            selector = lambda x: not x.hidden and x.group not in ('_Hidden',) and not x.internal_name.startswith('_') \
                                 and x.internal_name not in ('Edit', 'LinkTitleNoMenu', 'LinkTitle', 'DocIcon',
                                                             'ItemChildCount', 'FolderChildCount', 'AppAuthor',
                                                             'AppEditor', 'ComplianceAssetId', 'Modified', 'Created',
                                                             'Author', 'Editor', 'Attachments')
        return [i for i in self.get_list(name).fields.get().execute_query() if selector(i)]

    @property
    def users(self):
        if self._users is None:
            self._users = {}
            for i in self._context.web.site_users.get().execute_query():
                if i.principal_type == 1 and i.user_principal_name is not None:
                    name = i.title.split('(')[0].strip()
                    if ', ' in name:
                        family_name, given_name = name.split(', ')
                    else:
                        given_name, family_name = name.split('.')
                    given_name = given_name.capitalize()
                    family_name = family_name.capitalize()
                    name = f"{family_name}, {given_name}"
                    email = i.get_property('Email').lower()
                    self._users[i.id] = {'name': name, 'family_name': family_name, 'given_name': given_name,
                                         'email': email}
        return deepcopy(self._users)



import collections.abc
import threading

from ._item import item_factory


class List(collections.abc.Mapping):
    """
    Class representing Sharepoint List

    When inherited from, class attribute _title must be set.

    _map should be set, but can be overridden. This maps the python name (best to use a valid attribute-name)
    to the respective internal Sharepoint name. Do not map 'id' or 'ID', this is done automatically as the primary key.

    _data contains records (class Item) with the key being the sharepoint ID.

    At this point very optimistic (as in none-at-all) locking is used.

    The item itself keeps information if data has been changed.
    """
    _title = ''
    _lock = threading.Lock()
    _map = {}
    _upper_case = None
    _required = None

    @property
    def title(self) -> str:
        return self.__class__._title

    def __init__(self, sharepoint_site, lazy: bool = True, title: str = None, map: dict = None, upper_case = None, required = None):
        if not hasattr(self, '_sharepoint_site'):
            self._sharepoint_site = sharepoint_site
            self._title = title or self._title
            self._map = map or self._map
            self._upper_case = upper_case or self._upper_case
            self._required = required or self._required
            self._sharepoint_list = self._sharepoint_site.get_list(self.title)
            self._data = {}
            self._item_factory = item_factory(self)
        if not lazy:
            self.load()

    def load(self):
        for i in self._sharepoint_list.items.get().execute_query():
            item = self._item_factory(i)
            self._data[item.id] = item

    def rollback(self):
        self._data.clear()
        self.load()

    def get(self, item: int, buffered: bool = True):
        if not buffered or item not in self._data:
            item = self._item_factory(self._sharepoint_list.get_item_by_unique_id().execute_query())
            self._data[item.id] = item
        return self._data[id]

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __getitem__(self, item):
        return self._data[item]

    def commit(self):
        for i in self._data.values():
            i.commit(lazy=True)
        self._sharepoint_list.context.execute_batch()

    @property
    def python_map(self):
        return self._map.copy()

    @property
    def sharepoint_map(self):
        return {v: k for k, v in self._map.items()}

    def normalize(self):
        if self._upper_case:
            for i in self.values():
                for j in self._upper_case:
                    i[j] = i[j].upper()

    def validate(self):
        if self._required:
            errors = []
            for i in self.values():
                for j in self._required:
                    if not i[j]:
                        errors.append((i, j))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['base']

package_data = \
{'': ['*']}

install_requires = \
['Office365-REST-Python-Client>=2.3.12,<3.0.0',
 'click>=8.1.3,<9.0.0',
 'netlink-logging>=0.1.3,<0.2.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['sharepoint_list_info = '
                     'netlink.sharepoint.base.cli:print_list_info']}

setup_kwargs = {
    'name': 'netlink-sharepoint-base',
    'version': '0.0.0',
    'description': 'Basic Tools for interaction with Sharepoint Sites',
    'long_description': '# netlink-sharepoint-base\n\n## Basic Tools for interaction with Sharepoint Sites\n\nThis has been moved from `netlink-sharepoint`\n\nFor now only **Lists** are considered.\n\n## `sharepoint_list_info`\n\nScript to help mapping\n\n```shell\nUsage: sharepoint_list_info.py [OPTIONS] [NAME]...\n\n  Print information about SharePoint List(s)\n\n  NAME is not provided, all lists are returned.\n\nOptions:\n  -u, --url TEXT     Sharepoint URL\n  -i, --id TEXT      Client ID\n  -s, --secret TEXT  Client Secret\n  -t, --toml FILE    TOML file with \'url\', \'id\', and \'secret\'\n  -f, --fields       include fields\n  --hidden           include hidden lists\n```\n\n## `netlink.sharepoint.base.Site`\n\nMain class representing a Sharepoint site. Can either be used directly providing the parameters\n\n- url\n- client_id\n- client_secret\n\nor inherited from \n\n```python\nfrom netlink.sharepoint.base import Site as _Site\n\nclass Site(_Site):\n    _url = "https://somewhere.sharepoint.com/sites/something"\n    _id = "00000000-0000-0000-0000-000000000000"\n    _secret = "abcdefghijklmnopqrstuvwxyz01234567890+/abcd="\n```\n\n### url \n\n_read-only_\n\n### users\n\n_read-only (always a copy of the actual data)_\n\nDict of the users of the site.\n\n| Column | Type | Description |\n|--------|:-----:|-----|\n| id | int | reference key for AuthorID, EditorID, etc. |\n| family_name | str | a.k.a. last name |\n| given_name | str | a.k.a. first name |\n| name | str | {family_name}, {given_name} |\n| email | str | |\n\n**Note** This is based on a default Sharepoint setup. This might not work for you.\n\n### get_list(name)\n\nReturns a Sharepoint List object (`office365.sharepoint.lists.list.List`) by **name**.\n\n### get_lists(hidden=False)\n\nReturns a list of Sharepoint List objects (`office365.sharepoint.lists.list.List`). If `hidden` is `True`, internal Lists are included.\n\n### get_list_items(name)\n\nReturns list of all items (`office365.sharepoint.listitems.listitem.ListItem`) from **name**d Sharepoint List. \n\n### get_list_columns(name, hidden=False)\n\nReturns list of columns (a.k.a. Fields) of the **name**d Sharepoint List. If `hidden` is `True`, internal columns are included.\n\n### commit()\n\nSend all pending updates to Sharepoint.\n\n\n## `netlink.sharepoint.base.List`\n\nClass representing a Sharepoint List. This is a `collections.abc.Mapping`. Contents are mapped with the unique `id` of the item in the Sharepoint List.\n\nWhen inherited from, class attribute \n\n- `_title` must be set.\n\n- `_map` should be set, but can be overridden. This maps the python name (best to use a valid attribute-name)\n  to the respective internal Sharepoint name. Do not map \'id\' or \'ID\', this is done automatically as the primary key.\n  \n- `_upper_case` can be set to be used in `normalize` to enforce that the respective columns are set to uppercase.\n\n- `_required` can be set to ensure that the respective columns are not empty (as defined by Python: `None`, empty \n   string, or numeric zero).\n\nAt this point very optimistic (as in none-at-all) locking is used.\n\nThe item itself keeps information if data has been changed.\n\n### load()\n\nLoads all items from the Sharepoint List.\n\n### rollback()\n\nClears all local entries and calls `load`. At this time there is no difference, but when adding items will be supported, this would clear them.\n\n### get(item, buffered=True)\n\nLoads (if nor already loaded) item based on unique `id`. If `buffered` is `False`, read from Sharepoint is forced.\n\n### commit()\n\nCommit all changed items to Sharepoint.\n\n### normalize()\n\nProcesses `_upper_case` for all items.\n\nConsider overriding this to set columns based on other columns.\n\n### validate()\n\nProcesses `_required` for all items.\n\nConsider overriding this to check list-wide constraints, like composite unique keys. \n\n## Item\n\nThis is a special class that gets build dynamically for each List. It is a `collections.abc.MutableMapping`. \n\nIt handles keeping state of changes and processes the actual mapping between Python and Sharepoint columns (fields).\n\n### commit(lazy=False)\n\nUpdate the internal `office365.sharepoint.listitems.listitem.ListItem` with any pending changes. If `lazy` is `True`, changes are not sent to Sharepoint, providing the option to send on List or Site level.\n\n## License\n\nMIT\n',
    'author': 'Bernhard Radermacher',
    'author_email': 'bernhard.radermacher@netlink-consulting.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/netlink_python/netlink-sharepoint-base',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
